#!/usr/bin/env python

from comm import communicator, colorize
from pandas import read_hdf
from pandas import DataFrame, isnull
from tqdm import tqdm
import os
from datetime import *
from netCDF4 import Dataset
import glob
import h5py
import subprocess
import cPickle as pickle
from numpy import *

class obsdb(communicator):
    def __init__(self, **kwargs):
        communicator.__init__(self, kwargs.get('verbosity', 0))
        self.sites = DataFrame(
            columns=[
                'name',
                'tracer',
                'code',
                'mobile',
                'start','end',
                'platform',
                'lon','lat','alt','site_elevation'
                'backgroundFile','backgroundFile_valid','backgroundFile_hash'
                ]
        )
        self.observations = DataFrame(
            columns=[
                'site',
                'lat','lon','alt','time',
                'obs','err',
                'background','foreground',
                'model','mismatch',
                'file',
                'footprint', 'footprint_exists'
            ]
        )
        self.files = DataFrame(columns=['path'])
        self.filename = kwargs.get('filename', None)
        self.db_source = kwargs.get('copyfrom', None)
        self.new = self.db_source is None
        self.start = kwargs.get('tmin', None)
        self.end = kwargs.get('tmax', None)
        self.footprints_path = kwargs.get('footprints_path', None)

    def checkFootprints(self, path=None):
        self.footprints_path = path if path is not None else self.footprints_path
        if self.footprints_path is None :
            self.error("Unspecified footprints path")

        self.observations.loc[:, 'footprint'] = array(
            ['%s.%im.%s.h5'%(c.lower(), z, t.strftime('%Y-%m')) for (c, z, t) in zip(
                self.observations.site, self.observations.height, self.observations.time
            )]
        )
        self.observations.loc[:, 'footprint'] = [os.path.join(self.footprints_path, x) for x in self.observations.footprint]

        self.observations.loc[:, 'footprint_exists'] = False
        self.observations.loc[:, 'footprint_length'] = nan
        footprint_files = unique(self.observations.footprint)
        for fpf in tqdm(footprint_files, desc='Checking footprints'):
            if not os.path.exists(fpf):
                tqdm.write(colorize('<y>[WARNING] File <p:%s> not found! no footprints will be read from it</y>'%fpf))
            else :
                fp = h5py.File(fpf, mode='r')
                times = [x.to_pydatetime() for x in self.observations.loc[self.observations.footprint == fpf, 'time']]
                fp_exists = array([x.strftime('%Y%m%d%H%M%S') in fp for x in times])
                fp_exists[fp_exists] = [size(fp[x.strftime('%Y%m%d%H%M%S')].keys()) > 0 for x in array(times)[fp_exists]]
                self.observations.loc[self.observations.footprint == fpf, 'footprint_exists'] = fp_exists
                t1 = [sorted(fp[x.strftime('%Y%m%d%H%M%S')].keys())[0].split('_')[1] for x in self.observations.loc[self.observations.footprint == fpf].loc[fp_exists, 'time']]
                t2 = [sorted(fp[x.strftime('%Y%m%d%H%M%S')].keys())[-1].split('_')[0] for x in self.observations.loc[self.observations.footprint == fpf].loc[fp_exists, 'time']]
                t1 = [datetime.strptime(x, '%Y%m%d%H%M%S') for x in t1]
                t2 = [datetime.strptime(x, '%Y%m%d%H%M%S') for x in t2]
                dt = [tt2-tt1 for (tt2, tt1) in zip(t2, t1)]
                self.observations.loc[(self.observations.footprint == fpf) & (self.observations.footprint_exists), 'footprint_length'] = dt
        #
        #dbtemp = self.observations.loc[:, ['site',]]
        #dbtemp.loc[:, 'year'] = array([x.year for x in self.observations.loc[:, 'time']])
        #dbtemp.loc[:, 'month'] = array([x.month for x in self.observations.loc[:, 'time']])
        #dbtemp.loc[:, 'code'] = array([self.sites.loc[o.site].code.lower() for o in dbtemp.itertuples()])
        #dbtemp.loc[:, 'height'] = array([self.sites.loc[o.site].height for o in dbtemp.itertuples()])

        # Loop over each row of that database,
        #for oo in tqdm(dbtemp.itertuples(), desc='Searching footprint files', total=len(dbtemp)):

        #    # The file name should be code.year-month.h5, but for sites with multiple levels, it is codealt.year-month.h5
        #    fname = os.path.join(self.footprints_path,'%s.%04i-%02i.h5'%(oo.code, oo.year, oo.month))
        #    if not os.path.exists(fname): fname = fname.replace(oo.code, '%s.%im'%(oo.code, oo.height))

        #    # Write the footprint file name in the main database
        #    if os.path.exists(fname):
        #        selection = ((self.observations.site == oo.site) &
        #                     (self.observations.time >= datetime(oo.year, oo.month, 1)) &
        #                     (self.observations.time < datetime(oo.year, oo.month, 1) + relativedelta(months=1))
        #                    )
        #        self.observations.loc[selection, 'footprint'] = fname
        #    else :
        #        self.warn("File <p:%s> not found! no footprints will be read from it"%fname)

        ## Each file contains many footprints, here we compute the name of the groups containing the footprints for each obs
        #footprint_files = unique(self.observations.footprint)
        #footprint_files = footprint_files[~isnull(footprint_files)]
        #footprint_field = array([o.time.strftime('%Y%m%d%H%M%S') for o in self.observations.itertuples()])

        ## Open the files and look which fields are present
        #footprint_exists = zeros(len(footprint_field), dtype=bool)
        #nfiles = len(unique(footprint_files))
        #for ifile, file in tqdm(enumerate(unique(footprint_files)), total=nfiles, desc="Checking footprint files"):
        #    indices = self.observations.footprint == file
        #    ds = h5py.File(file, 'r')
        #    footprint_exists[indices] = array([x in ds.keys() for x in footprint_field[indices]])
        #    ds.close()

        ## Save that to the database
        ## self.db.observations.loc[footprint_exists, 'footprint'] = footprint_files[footprint_exists]
        #self.observations.loc[:, 'footprint_exists'] = footprint_exists

    def migrateFootprints(self, cache):
        exist = self.observations.loc[:, 'footprint_exists']
        names = unique(self.observations.loc[exist, 'footprint'])
        for old in tqdm(names, desc="Copy footprint files to cache"):
            new = old.replace(self.footprints_path, cache)
            if not os.path.exists(os.path.dirname(new)):
                os.makedirs(os.path.dirname(new))
            subprocess.check_call(['rsync', old, new])
            self.observations.loc[self.observations.footprint == old, 'footprint'] = new

    def readBackgrounds(self, backgroundFile):
        from Tools import interpTime
        with open(backgroundFile, 'r') as fid :
            bg = pickle.load(fid)
        times = bg['time']
        tt1 = array([(t-amin(times)).total_seconds() for t in times])
        for isite, site in enumerate(tqdm(self.sites.itertuples(), total=len(self.sites))):
            index_obs = self.observations.site == site.Index
            times_obs = self.observations.loc[index_obs, 'time']
            try :
                try :
                    data = bg[site.code]
                except KeyError:
                    data = bg['%s%03i'%(site.code, site.height)]
                tt0 = array([(t-amin(times)).total_seconds() for t in times_obs])
                self.observations.loc[index_obs, 'background'] = interp(tt0, tt1, data['bgd']) #interpTime(times_obs, times, data['bgd'])
                self.observations.loc[index_obs, 'TM5'] = interp(tt0, tt1, data['tm5']) #interpTime(times_obs, times, data['tm5'])
            except KeyError :
                self.warn("No background found for site %s (%s)"%(site.name, site.code))
        del bg

    def setMinErr(self, errmin):
        self.observations.loc[self.observations.err < errmin, 'err'] = errmin

    def importFromObspack(self, path, pattern='co2_*.nc'):
        sites = []
        #observations = DataFrame(columns = ['site', 'lat', 'lon', 'alt', 'time', 'obs', 'err', 'file'])
        tmin = datetime(1000,1,1)
        tmax = datetime(3000,1,1)
        if self.start is not None : tmin = self.start
        if self.end is not None : tmax = self.end

        for file in sorted(glob.glob(os.path.join(path, pattern))) :
            self.files = self.files.append(DataFrame.from_dict({'path':[file]}), ignore_index=True)
            ds = Dataset(file, 'r')
            scale = 1.
            if ds.dataset_parameter == 'co2' : scale = 1.e6
            if ds.dataset_parameter == 'ch4' : scale = 1.e9
            elevation = ds.site_elevation
            time = array([datetime(*x) for x in ds.variables['time_components'][:]])
            selection = (time >= tmin)*(time < tmax)
            nobs = len(time[selection])
            try :
                if 'value_unc' in ds.variables.keys():
                    err = ds.variables['value_unc'][selection]*scale
                else :
                    err = zeros(sum(selection))
                data = {
                    'time':time[selection],
                    'lat':ds.variables['latitude'][selection],
                    'lon':ds.variables['longitude'][selection],
                    'alt':ds.variables['altitude'][selection],
                    'obs':ds.variables['value'][selection]*scale,
                    'err':err,
                    'file':zeros(nobs)+self.files.loc[self.files.path == file].index[0],
                    'site':zeros(nobs)-1
                }
                sampling_heights = unique(data['alt']-ds.site_elevation)
            except :
                import pdb; pdb.set_trace()
            
            for sampling_height in sampling_heights :
                try :
                    sitename = ds.site_name
                except :
                    sitename = ds.site_code
                try :
                    site = {
                        'code':ds.site_code,
                        'name':sitename,
                        'alt':ds.site_elevation,
                        'lat':ds.site_latitude,
                        'lon':ds.site_longitude,
                        'height':sampling_height
                    }
                except :
                    import pdb; pdb.set_trace()
                if not site in sites :
                    sites.append(site)
                data['site'][data['alt']-ds.site_elevation == sampling_height] = sites.index(site)

            ds.close()
            self.msg("<i>%6i</i> observations imported from file <p:%s>"%(nobs, file))
            self.observations = self.observations.append(DataFrame.from_dict(data))

        self.observations = self.observations.reset_index(drop=True)
        self.sites = DataFrame.from_dict(sites)

        duplicates = set()
        for site in self.sites.itertuples():
            if site.code in duplicates or count_nonzero(self.sites.code == site.code) > 1 :
                duplicates.add(site.code)
                self.sites.loc[site.Index, 'code'] = '%s%03i'%(site.code, site.height)
        
    def selectByRegion(self, region):
        self.observations = self.observations.loc[
            (self.observations.lon >= region.lonmin) &
            (self.observations.lon <= region.lonmax) &
            (self.observations.lat >= region.latmin) &
            (self.observations.lat <= region.latmax)
        ]
        #TODO: need for a cleanup method for sites and footprint tables

    def SelectByTime(self, start, end):
        self.observations = self.observations.loc[(
            (self.observations.time >= start) &
            (self.observations.time <= end)
        )]
        self.sites = self.sites.loc[unique(self.observations.site), :]

    # def attribBackgroundFiles(self, path):
    #     self.msg("Searching for background files in path: %s"%path)
    #     id, fnames = [], []
    #     for site in self.sites.itertuples():
    #         fnames.append(os.path.join(path, '%s.%s.bgd.nc4')%(site.tracer.lower(), site.code))
    #         id.append(site.Index)
    #     self.sites.loc[id, 'backgroundFile'] = fnames
    #     self.sites.loc[id, 'backgroundFile_valid'] = [os.path.exists(x) for x in fnames]
    #     for site in self.sites.loc[~self.sites.backgroundFile_valid].itertuples() :
    #         self.warn("No background file found for site %s"%siteinfo(site))
    #
    # def readBackgrounds(self, path):
    #     self.attribBackgroundFiles(path)
    #     for site in tqdm(self.sites.query('backgroundFile_valid == True').itertuples()):
    #         bgd = backgroundFile(site.backgroundFile)
    #         hash = md5(site.backgroundFile).hexdigest()
    #         if hash != site.backgroundFile_hash :
    #             times_obs = self.observations.loc[self.observations.site == site.Index, 'time']
    #             if len(times_obs) > 0 :
    #                 if not bgd.coversPeriod(times_obs.min(), times_obs.max()):
    #                     self.sites.loc[site.Index, 'backgroundFile_valid'] = False
    #                     self.warn("The background file %s does not cover the whole data period (%s to %s)"%(site.backgroundFile, times_obs.min().strftime('%c'), times_obs.max().strftime('%c')))
    #                     self.warn("No background values will be interpolated for this site:")
    #                     self.warn(siteinfo(site))
    #                     self.warn('Background file:')
    #                     self.warn(site.backgroundFile)
    #                 else :
    #                     background_interp = bgd.interp('mixing_ratio_background', times_obs)
    #                     self.observations.loc[self.observations.site == site.Index, 'background'] = background_interp
    #                     self.sites.loc[site.Index, 'backgroundFile_hash'] = md5(site.backgroundFile).hexdigest()
    #                     self.sites.loc[site.Index, 'backgroundFile_valid'] = True

    def from_pickle(self, filename):
        import cPickle
        file = open(filename, 'r')
        db = cPickle.load(file)
        file.close()
        self.observations = DataFrame.from_dict(db['observations'])
        self.sites = DataFrame.from_dict(db['sites'])
        if 'path' in db['files']:
            self.files = DataFrame.from_dict(db['files'])
        else :
            self.diles = DataFrame.from_dict({'path':db['files']})
        if self.start is not None and self.end is not None :
            self.SelectByTime(self.start, self.end)
        else :
            self.start = db['tmin']
            self.end = db['tmax']
        del db

    def from_hdf(self, filename):
        df = read_hdf(filename)
        df.loc[:, 'site'] = df.siteCode
        df.loc[:, 'obs'] = df.mix*1.e6
        df.loc[:, 'err'] = df.err*1.e6
        self.observations = df.loc[:,['time', 'lat', 'lon', 'alt', 'height', 'obs', 'err', 'site', 'footprint', 'file']]

        # sites database:
        sites = df.loc[:, ['siteName', 'siteCode', 'eurocomCode', 'lon', 'lat', 'alt', 'height']].drop_duplicates()
        sites.loc[:, 'code'] = sites.siteCode
        sites.loc[:, 'name'] = sites.siteName
        self.sites = sites.loc[:, ['alt', 'lat', 'lon', 'height', 'code', 'name', 'siteCode']]
        self.sites.set_index('siteCode', inplace=True)

        # Files database:
        self.files = df.file.drop_duplicates()
        for file in self.files.iteritems():
            self.observations.loc[self.observations.file == file[1], 'file'] = file[0]

    def to_h5(self, filename=None):
        import h5py
        def tableToGroup(table, group):
            dtvars = ''
            for column in table.columns:
                if issubdtype(table.loc[:, column], 'datetime64'):
                    group[column] = array([x.timetuple()[:6] for x in table.loc[:, column]])
                    dtvars += column+';'
                else :
                    group[column] = table.loc[:, column]
            group['index'] = array([x.Index for x in table.itertuples()])
            group.attrs['datetime columns'] = dtvars.strip(';')
            return

        if filename is None: filename = 'observations.h5'
        ds = h5py.File(filename)
        ds.create_group('observations')
        ds.create_group('sites')
        ds.create_group('files')
        tableToGroup(self.observations, ds['observations'])
        tableToGroup(self.sites, ds['sites'])
        tableToGroup(self.files, ds['files'])
        ds.close()

    def write(self, filename=None):
        import cPickle
        db = dict()
        db['observations'] = self.observations.to_dict()
        db['sites'] = self.sites.to_dict()
        db['files'] = self.files.to_dict()
        db['tmin'] = self.start
        db['tmax'] = self.end
        if filename is None : filename = self.filename
        if filename is not None :
            if not os.path.exists(os.path.dirname(filename)):
                if os.path.dirname(filename) != '' :
                    os.makedirs(os.path.dirname(filename))
            file = open(filename, 'w')
        else :
            self.error("Please provide a file name")
        self.msg("Writing observation database to <p:%s>"%file.name)
        cPickle.dump(db, file, cPickle.HIGHEST_PROTOCOL)
        file.close()

    def write_hdf(self, filename=None):
        if filename is None : filename = self.filename
        if filename is not None :
            if not os.path.exists(os.path.dirname(filename)):
                if os.path.dirname(filename) != '' :
                    os.makedirs(os.path.dirname(filename))
            file = open(filename, 'w')
        else :
            self.error("Please provide a file name")
        self.msg("Writing observation database to <p:%s>"%file.name)
        self.observations.to_hdf(filename, 'observations')
        self.sites.to_hdf(filename, 'sites')
        self.files.to_hdf(filename, 'files')

    def read_hdf(self, filename):
        self.observations = read_hdf(filename, 'observations')
        self.sites = read_hdf(filename, 'sites')
        self.files = read_hdf(filename, 'files')
        
    def append(self, dbf):
        db2 = obsdb()
        db2.read_hdf(dbf)
        self.observations = self.observations.append(db2.observations, ignore_index=True).drop_duplicates()
        self.sites = self.sites.append(db2.sites).drop_duplicates()
        self.files = self.files.append(db2.files).drop_duplicates()
        del db2

    def selectMinDaily(self):
        indices = []
        for site in self.sites.itertuples():
            selection = self.observations.loc[self.observations.site == site.Index, :]
            days = array([x.date() for x in selection.loc[:, 'time']])
            for day in unique(days):
                indices.append(selection.loc[days == day, 'obs'].argmin())
        return indices

    def reduceFreq(self, maxfreq=3):
        """
        Ensure that we don't mave more than one obs every "maxfreq" at each site
        Loop over the obs and pick the one that has the lowest uncertainty in each maxfreq interval, for every site
        """
        dt = timedelta(hours=maxfreq)
        tokeep = []
        for site in tqdm(self.sites.itertuples(), desc="Reduce observation frequency to a max of %i obs/hour"%maxfreq, total=self.sites.shape[0]):
            db = self.observations.loc[self.observations.site == site.Index]
            tt = datetime(*db.time.min().timetuple()[:3])
            tmax = datetime(*db.time.max().timetuple()[:6])
            pb = tqdm(total=ceil((tmax-tt).total_seconds()/3600./maxfreq), leave=False)
            while tt < tmax :
                obs = db.loc[(db.time >= tt) & (db.time < tt+dt), 'err'].dropna()
                if obs.shape[0] > 0 :
                    try :
                        tokeep.append(obs.idxmin())
                    except :
                        import pdb; pdb.set_trace()
                tt += dt
                pb.update()
            pb.close()
        self.reduce(tokeep)

    def getIndicesAtSites(self, sitecodes, invert=False):
        indices = []
        for site in tqdm(sitecodes, leave=False) :
            if site in self.sites.code.values :
                site_index = self.sites.loc[self.sites.code == site].index[0]
                indices.extend(self.observations.loc[self.observations.site == site_index].index.tolist())
        if invert :
            idx = [x for x in self.observations.index if not x in array(indices)]
            return idx
        return indices

    def reduce(self, indices):
        self.observations = self.observations.loc[indices]
        self.sites = self.sites.loc[unique(self.observations.site), :]

    def excludeSites(self, sitelist):
        indices = self.getIndicesAtSites(sitelist)
        self.observations = self.observations.drop(indices)
        self.sites = self.sites.loc[unique(self.observations.site), :]

    def excludeFromList(self, excludelist):
        selection = self.parse_config_file(excludelist)

        # Exclude sites
        excluded = [x for x in selection.keys() if not selection[x]['selected']]
        self.excludeSites(excluded)

        # Limit the time period of some time series
        for site in tqdm(self.sites.itertuples(), total=self.sites.shape[0], leave=False) :
            if selection[site.code]['restricted'] is not None :
                db2 = self.observations.loc[
                    (self.observations.site == site.Index) &
                    (self.observations.time >= selection[site.code]['restricted'][0]) &
                    (self.observations.time < selection[site.code]['restricted'][1])
                ]
                self.observations = self.observations.loc[self.observations.site != site.Index]
                self.observations = self.observations.append(db2)

        # Filter day/night time observations :
        for site in tqdm(self.sites.itertuples(), total=self.sites.shape[0], leave=False) :
            size0 = self.observations.shape[0]
            interval = selection[site.code]['interval']
            if interval != 'all':
                db_site = self.observations.loc[self.observations.site == site.Index]
                if '{' in interval and '}' in interval :
                    intervals = interval.strip('{').strip('}').split()
                    for interval in intervals :
                        mrange, interval = interval.split(':')
                        if interval != 'all' :
                            m1, m2 = [datetime.strptime(x, '%Y%m') for x in mrange.split('-')]
                            i1, i2 = [int(x) for x in interval.split('-')]
                            db_range = db_site.loc[(db_site.time >= m1) & (db_site.time < m2)]
                            db_site = db_site.loc[(db_site.time < m1) | (db_site.time >= m2)]
                            hours = array([x.hour for x in db_range.time])
                            if i2 > i1 :
                                db_range = db_range.iloc[(hours >= i1) & (hours < i2)]
                            else :
                                db_range = db_range.iloc[(hours < i2) | (hours >= i1)]
                    db_site = db_site.append(db_range)
                else :
                    i1, i2 = [int(x) for x in interval.split('-')]
                    hours = array([x.hour for x in db_site.time])
                    if i2 > i1 :
                        db_site = db_site.iloc[(hours >= i1) & (hours < i2)]
                    else :
                        db_site = db_site.iloc[(hours < i2) | (hours >= i1)]
                self.observations = self.observations.loc[self.observations.site != site.Index]
                self.observations = self.observations.append(db_site)

        # Eliminate the empty sites from the database
        self.sites = self.sites.loc[unique(self.observations.site), :]

    def parse_config_file(self, file):
        sites = {}
        fid = open(file, 'r')
        for line in fid.readlines():
            line = line.split('#')[0]
            fields = [x.strip() for x in line.split(';')]
            sites[fields[0]] = {
                'interval' : fields[1],
                'selected' : True,
                'restricted' : None
            }
            if len(fields) > 2 :
                if fields[2] == 'X' :
                    sites[fields[0]]['selected'] = False
            if len(fields) > 3 :
                try :
                    interval = fields[3].replace('S', self.start.strftime('%Y%m%d')).replace('E', self.end.strftime('%Y%m%d'))
                    s, e = interval.split('-')
                    sites[fields[0]]['restricted'] = (datetime.strptime(s, '%Y%m%d'), datetime.strptime(e, '%Y%m%d'))
                except :
                    if not x.startswith('!'): print line
        fid.close()
        return sites


def siteinfo(site):
    if site.mobile :
        return '%s (%s) : mobile site (%s)'%(site.code, site.name, site.platform)
    else :
        return '%s (%s) : %.1fN, %.1fE, %i m a.m.s.l'%(site.code, site.name, site.lat, site.lon, site.alt)


class backgroundFile(communicator):
    def __init__(self, filename):
        super(backgroundFile, self).__init__()
        self.filename = filename
        try :
            ds = Dataset(self.filename, 'r')
        except :
            self.error("File %s not found"%self.filename)
        self.times = array([datetime(*x) for x in ds.variables['time'][:]])
        dt = self.times[1]-self.times[0]
        self.start = self.times.min()-dt/2
        self.end = self.times.max()+dt/2
        ds.close()

    def coversPeriod(self, ti, tf):
        if ti < self.start or tf > self.end :
            return False
        else :
            return True

    def interp(self, field, time):
        from numpy import interp, amin
        ds = Dataset(self.filename, 'r')
        data = ds.variables[field][:]
        ds.close()
        tref = min(amin(self.times), amin(time))
        tInt = array([(x-tref).total_seconds() for x in self.times])
        tExt = array([(x-tref).total_seconds() for x in time])
        return interp(tExt, tInt, data)

class profileFile(communicator):
    def __init__(self, filename):
        super(profileFile, self).__init__()
        self.filename = filename
        try :
            ds = Dataset(self.filename, 'r')
        except :
            self.error("File %s couldn't be opened"%self.filename)
        self.times = array([datetime(*x) for x in ds['date_midpoints']])
        self.profile = ds['mix'][:]
        self.times = self.times[self.profile[0,0,:] > 0]
        self.times_int = array([(x-self.times[0]).total_seconds() for x in self.times])
        self.profile = self.profile[:,:,self.profile[0,0,:]>0]
        self.levels = ds['levels'][:]
        self.co2 = None
        self.co2_bg = None
        ds.close()

    def sample(self, times, height):
        times = array([x.to_pydatetime() for x in times])
        self.sample_at_height(height)
        co2 = self.sample_at_time(times, 'co2')
        co2_bg = self.sample_at_time(times, 'co2_bg')
        return co2, co2_bg

    def sample_at_height(self, height):
        l1 = next(x for (x,v) in enumerate(self.levels) if v >= height)
        l0 = l1-1 if l1 > 0 else 0
        d0, d1, dtot = height-self.levels[l0], self.levels[l1]-height, self.levels[l1]-self.levels[l0]
        r1, r0 = d0/dtot, d1/dtot
        self.co2 = self.profile[l0, 0, :]*r0 + self.profile[l1, 0, :]*r1
        self.co2_bg = self.co2-self.profile[l0, 1, :]*r0 + self.profile[l1, 1, :]*r1

    def sample_at_time(self, times, tracer='co2_bg'):
        tref = self.times[0]
        times_ext = array([(x-tref).total_seconds() for x in times])
        return interp(times_ext, self.times_int, getattr(self, tracer))
