#!/usr/bin/env python
from pandas import DataFrame, read_hdf
from optimizer import Categories
from netCDF4 import Dataset
#from numpy import array, dot, append, zeros_like, unique, zeros, eye, exp, ones, pi, sin, cos, arcsin, transpose, meshgrid, arange, linalg, argsort, where, diag, flipud, sqrt, shape, amax, abs
from numpy import *
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from tqdm import tqdm
#from lucdas.tm5_utils import precon
import os
from Tools import Region
from comm import communicator
import h5py
from lumia.rctools import rc

class optimData(communicator):
    def __init__(self, rcf=None, verbosity=0, prior=None, savefile=None):
        self.model_data = {}
        communicator.__init__(self, verbosity)
        self.horizontal_correlations = {}
        self.temporal_correlations = {}
        self.vectors = DataFrame(columns=[
            'time_start',
            'time_end',
            'lat',
            'lon',
            'state_prior',
            'state_prior_preco',
            'state_posterior',
            'prior_uncertainty',
            'posterior_uncertainty',
            'category',
            'land_fraction'
        ])
        if savefile is None :
            self.rcf = rcf
            self.categories = Categories(rcf)
        else :
            self.load(savefile)
        self.region = Region(self.rcf)
        self.start = datetime(*self.rcf.get('time.start'))
        self.end = datetime(*self.rcf.get('time.end'))
        if prior is not None : self.fillEmis(prior)

    def readEmis(self, emfile):
        """
        Import the emissions from a previous run
        #TODO: this class should be the counterpart of one class that writes the optimized state
        """
        with Dataset(emfile) as ds :
            categories = ds.groups.keys()
            data = {'cat_list':categories}
            for cat in categories :
                tstart = array([datetime(*x) for x in ds[cat]['times_start'][:]])
                tend = array([datetime(*x) for x in ds[cat]['times_end'][:]])
                select_time = (tstart >= self.start)*(tend <= self.end)
                if tstart[select_time].min() > self.start or tend[select_time].max() < self.end :
                    msg = "Mismatch between the requested emissions and the emissions in the file <p:%s>"%emfile
                    msg += "For the category <b:%s>:"%cat
                    msg += "Period requested from %s to %s"%(self.start.strftime('%d %b %Y %H:%M'), self.end.strftime('%d %B %Y %H:%M'))
                    msg += "Period found from %s to %s"%(tstart.min().strftime('%d %b %Y %H:%M'), tend.max().strftime('%d %B %Y %H:%M'))
                    self.error(msg)
                data[cat] = {
                    'emis':ds[cat]['emis'][select_time, :, :],
                    'time_interval': {
                        'time_start': tstart[select_time],
                        'time_end': tend[select_time]
                    }
                }
        self.fillEmis(data)
        #self.model_data['prior_emis'] = data
        #self.fillVectors(data)

    def fillEmis(self, emstruct):
        self.setPrior(emstruct)

    def setPrior(self, emstruct):
        self.model_data['prior_emis'] = emstruct
        self.fillVectors(emstruct)

    def fillVectors(self, emis):
        """
        This method creates a state vector (and auxiliary variables such as coordinates and category
        index from an emission structure
        :param emis:
        :return:
        """
        lsm = self.region.get_land_mask(refine_factor=2)

        # Containers
        statevec_std = []
        statevec_min = []
        statevec_max = []
        statevec = []
        lat, lon, cats, land_fraction = [], [], [] ,[]
        time_start, time_end = array(()), array(())

        for cat in [x for x in self.categories if x.optimize]:
            time = []
            emcat, emcat_std = coarsenTime(emis[cat.name], cat.optimization_interval, compute_std=True)
            for i_time, tt in enumerate(sorted(emcat['time_interval']['time_start'])):
                for i_lat in xrange(emcat['emis'].shape[1]):
                    for i_lon in xrange(emcat['emis'].shape[2]):
                        statevec.append(emcat['emis'][i_time, i_lat, i_lon])
                        statevec_std.append(emcat_std['std'][i_time, i_lat, i_lon])
                        statevec_min.append(emcat_std['min'][i_time, i_lat, i_lon])
                        statevec_max.append(emcat_std['max'][i_time, i_lat, i_lon])
                        lon.append(self.region.lons[i_lon])
                        lat.append(self.region.lats[i_lat])
                        time.append(tt)
                        cats.append(cat.name)
                        land_fraction.append(lsm[i_lat, i_lon])

            dt = {'y':relativedelta(years=1), 'm':relativedelta(months=1), 'd':timedelta(1)}
            dt = dt[cat.optimization_interval]
            time_start = append(time_start, array(time))
            time_end = append(time_end, array(time)+dt)

        self.vectors.loc[:, 'lon'] = lon
        self.vectors.loc[:, 'lat'] = lat
        self.vectors.loc[:, 'time_start'] = time_start
        self.vectors.loc[:, 'time_end'] = time_end
        self.vectors.loc[:, 'category'] = cats
        self.vectors.loc[:, 'land_fraction'] = land_fraction
        self.vectors.loc[:, 'state_prior'] = statevec
        self.vectors.loc[:, 'state_prior_std'] = statevec_std
        self.vectors.loc[:, 'state_prior_min'] = statevec_min
        self.vectors.loc[:, 'state_prior_max'] = statevec_max
        self.vectors.loc[:, 'state_prior_preco'] = 0

    def setupUncertainties(self):
        self.calcPriorUncertainties()
        self.setup_Hcor()
        self.setup_Tcor()

    def calcPriorUncertainties(self):
        """
        This method computes the diagonal elements of the prior covariance matrix.
        For the moment, only a percentage of the prior is implemented, but more advanced
        uncertainty calculations should be implemented here
        :return: a state-vector like vector of uncertainties.
        """
        #TODO: structToState should be just for one category
        #TODO: implement the land/sea masks in a better way? (urgent!)
        #TODO: provide the uncertainty externally

        for cat in self.categories :
            if cat.optimize :
                errfact = cat.uncertainty*0.01
                # Uncertainty can be with defined as a fraction of the total, or with reference to
                # the temporal variability within each optimized interval
                if cat.uncertainty_type == 'tot' :
                    errcat = abs(self.vectors.loc[self.vectors.category == cat, 'state_prior'].values)*errfact
                elif cat.uncertainty_type == 'sig' :
                    errcat = abs(self.vectors.loc[self.vectors.category == cat, 'state_prior_std'].values)*errfact
                else :
                    raise NotImplemented

                # Apply a minimum uncertainty if needed:
                min_unc = cat.min_uncertainty*errcat.max()/100.
                #errcat[errcat < cat.min_uncertainty] = cat.min_uncertainty
                #errcat[errcat < min_unc] = min_unc

                # Note 20 feb 2019:
                # No land filter needed as long as the state variables are in units of 
                # nt.umol/m2/sec, but land filter still needed to avoid setting non-zero
                # uncertainties to ocean
                #
                # Land filter (avoid modifying biosphere over ocean and conversely ...):
                land_filter = self.vectors.land_fraction.values
                errcat[(errcat < min_unc) & (land_filter > 0)] = min_unc
                #if cat.is_ocean : land_filter = 1-land_filter
                #errcat *= land_filter 

                self.vectors.loc[self.vectors.category == cat, 'prior_uncertainty'] = errcat

    def setup_Hcor(self):
        for cat in self.categories :
            if cat.optimize :
                if not cat.horizontal_correlation in self.horizontal_correlations :
                    fname = self.checkCorFile(cat.horizontal_correlation, cat) # TODO: Move this to a completely external code/module?
                    P_h, D_h = read_latlon(fname)
                    Hor_L = P_h * D_h
                    self.horizontal_correlations[cat.horizontal_correlation] = Hor_L
                    del P_h, D_h

    def checkCorFile(self, hcor, cat):
        # Generate the correlation file name
        corlen, cortype = hcor.split('-')
        corlen = int(corlen)
        fname = 'Bh:%s:%5.5i_%s.nc'%(
            self.region.name,
            corlen,
            cortype
        )
        fname = os.path.join(self.rcf.get('correlation.inputdir'), fname)
        if not os.path.exists(fname):
            self.msg("Correlation file <p:%s> not found. Computing it"%fname)
            hc = horcor(
                self.region,
                corlen,
                cortype,
                self.vectors.loc[self.vectors.category == cat,('lat', 'lon')].drop_duplicates()
            )
            hc.calc_latlon_covariance()
            hc.write(fname)
            del hc
        return fname

    def setup_Tcor(self):
        for cat in self.categories :
            if cat.optimize :
                if not cat.temporal_correlation in self.temporal_correlations :
                    temp_corlen = float(cat.temporal_correlation[:3].strip())

                    # Time interval of the optimization
                    dt = {'y':12., 'm':1, 'd':1/30.}[cat.optimization_interval]

                    # Number of time steps :
                    times = self.vectors.loc[self.vectors.category == cat, 'time_start'].drop_duplicates()
                    nt = times.shape[0]

                    P_t, D_t = calc_temp_corr(temp_corlen, dt, nt)
                    self.temporal_correlations[cat.temporal_correlation] = dot(P_t, D_t)
        #self.writeTemporalCorrelations()  # why would we write that?

    def xc_to_x(self, state_preco, add_prior=True):
        uncertainty = self.vectors.loc[:, 'prior_uncertainty'].values
        state = 0*uncertainty
        catIndex = self.vectors.category.tolist()
        for cat in tqdm(self.categories) :
            if cat.optimize :
                Hor_L = self.horizontal_correlations[cat.horizontal_correlation]
                Temp_L = self.temporal_correlations[cat.temporal_correlation]
                ipos = catIndex.index(cat.name)
                state += xc_to_x(uncertainty, Temp_L, Hor_L, state_preco, ipos, 1)
        if add_prior: state += self.vectors.loc[:, 'state_prior']
        return state


    def g_to_gc(self, g):
        g_c = zeros_like(g)
        state_uncertainty = self.vectors.loc[:, 'prior_uncertainty'].values
        catIndex = self.vectors.category.tolist()
        for cat in tqdm(self.categories):
            if cat.optimize :
                Hor_Lt = self.horizontal_correlations[cat.horizontal_correlation].transpose()
                Temp_Lt = self.temporal_correlations[cat.temporal_correlation].transpose()
                ipos = catIndex.index(cat.name)
                g_c += g_to_gc(state_uncertainty, Temp_Lt, Hor_Lt, g, ipos, 1)
        return g_c

    def store(self, var, field, filter=None):
        if filter is None :
            self.vectors.loc[:, field] = var
        else :
            self.vectors.loc[filter, field] = var

    def read(self, field, filter=None):
        if filter is None :
            return self.vectors.loc[:, field]
        else :
            return self.vectors.loc[filter, field]

    def __len__(self):
        return self.vectors.shape[0]

    def __getattr__(self, item):
        if item is 'size' :
            return len(self)
        else :
            if hasattr(self, item):
                return getattr(self, item)
            else :
                raise AttributeError(item)

    def save(self, filename):
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))
        self.vectors.to_pickle(filename)

    def write(self, filename):
        if os.path.exists(filename):
            os.remove(filename)
        self.vectors.to_hdf(filename, 'vectors')
        fid = h5py.File(filename, 'a')

        # Auxiliary data
        aux = fid.create_group('model_data')
        for cat in self.categories :
            grp = aux.create_group(cat.name)
            grp['apri'] = self.model_data['prior_emis'][cat.name]['emis']
            grp['time_start'] = array([x.timetuple()[:6] for x in self.model_data['prior_emis'][cat.name]['time_interval']['time_start']])
            grp['time_end'] = array([x.timetuple()[:6] for x in self.model_data['prior_emis'][cat.name]['time_interval']['time_end']])
            for attr in cat.__dict__:
                grp.attrs[attr] = getattr(cat, attr)

        # Region :
        fid['lats'] = self.region.lats
        fid['lons'] = self.region.lons
        fid.attrs['region'] = self.region.name

        # Correlations
        corr = fid.create_group('correlations')
        hc = corr.create_group('hor')
        for cor in self.horizontal_correlations :
            hc[cor] = self.horizontal_correlations[cor]
        tc = corr.create_group('temp')
        for cor in self.temporal_correlations :
            tc[cor] = self.temporal_correlations[cor]

        # rcf :
        rcf = fid.create_group('rcf')
        for key in self.rcf.keys :
            rcf.attrs[key] = self.rcf.get(key)

        fid.close()

    def load(self, filename, loadrc=True):
        """
        Counterpart of write
        #TODO : missing the setup of categories and region from the save file directly
        #TODO : read/save the rc keys as expressions and not values
        """
        self.vectors = read_hdf(filename, 'vectors')
        fid = h5py.File(filename, 'r')

        # Auxiliary data
        if loadrc :
            self.rcf = rc()
            for key in fid['rcf'].attrs:
                self.rcf.setkey(key, fid['rcf'].attrs[key])
            self.categories = Categories(self.rcf)

        self.model_data['prior_emis'] = {'cat_list':self.categories.list}
        for cat in self.categories :
            aux = fid['model_data'][cat.name]
            self.model_data['prior_emis'][cat.name] = {
                'emis': aux['apri'][:],
                'time_interval':{
                    'time_start': array([datetime(*x) for x in aux['time_start'][:]]),
                    'time_end': array([datetime(*x) for x in aux['time_end'][:]])
                }
            }
            for attr in aux.attrs :
                setattr(cat, attr, aux.attrs[attr])

        for cor in fid['correlations/hor']:
            self.horizontal_correlations[cor] = fid['correlations/hor'][cor][:]
        for cor in fid['correlations/temp']:
            self.temporal_correlations[cor] = fid['correlations/temp'][cor][:]

        fid.close()

    def reduce_time_span(self, tmin=None, tmax=None):
        """
        Reduces the size of the state vector by reducing the time span
        """
        # 1) Reduces the "vectors" object
        self.vectors = self.vectors.loc[(self.vectors.time_start >= tmin) & (self.vectors.time_end <= tmax), :]
        self.vectors.reset_index(inplace=True)

        # 2) Reduces the auxiliary data
        for cat in self.categories:
            dat = self.model_data['prior_emis'][cat.name]
            s1 = dat['time_interval']['time_start'] >= tmin
            s2 = dat['time_interval']['time_end'] <= tmax
            dat['emis'] = dat['emis'][s1*s2, :, :]
            dat['time_interval']['time_start'] = dat['time_interval']['time_start'][s1*s2]
            dat['time_interval']['time_end'] = dat['time_interval']['time_end'][s1*s2]

        # 3) Recompute the temporal correlations:
        self.temporal_correlations = {}
        self.setup_Tcor()

        # 4) fix the start/end attributes
        self.start = tmin
        self.end = tmax

def coarsenTime(emis, interval, compute_std=False):
    intervals_emis = emis['time_interval']['time_start']
    if interval == 'y' :
        intervals_optim = [datetime(x.year, 1, 1) for x in intervals_emis]
        dt = relativedelta(years=1)
    elif interval == 'm':
        intervals_optim = [datetime(x.year, x.month, 1) for x in intervals_emis]
        dt = relativedelta(months=1)
    elif interval == 'd' :
        intervals_optim = [datetime(x.year, x.month, x.day) for x in intervals_emis]
        dt = timedelta(1)
    else :
        raise NotImplementedError
    emisOut = []
    emisOut_std = []
    emisOut_min = []
    emisOut_max = []
    intervals_optim = array(intervals_optim)
    for tt in unique(intervals_optim):
        try :
            emisOut.append(emis['emis'][intervals_optim == tt, :, :].sum(0))
        except :
            import pdb; pdb.set_trace()
        nt = sum(intervals_optim == tt)
        if compute_std :
            emisOut_std.append(emis['emis'][intervals_optim == tt, :, :].std(0)*nt)
            emisOut_min.append(emis['emis'][intervals_optim == tt, :, :].min(0)*nt)
            emisOut_max.append(emis['emis'][intervals_optim == tt, :, :].max(0)*nt)
    em = {'time_interval':{}}
    em['emis'] = array(emisOut)
    em['time_interval']['time_start'] = unique(intervals_optim)
    em['time_interval']['time_end'] = unique(intervals_optim)+dt
    if compute_std :
        return em, {'std':array(emisOut_std), 'min':array(emisOut_min), 'max':array(emisOut_max)}
    return em


class horcor:
    def __init__(self, region, corlen, cortype, statevec, min_eigval=0.00001):
        self.corlen = corlen
        self.cortype = cortype
        self.state = statevec
        self.min_eigval = min_eigval
        self.region = region
        self.n_hor = -1
        self.lam = None
        self.P = None
        self.P_diag = None

    def calc_latlon_covariance(self):
        from numpy.linalg import eigh
        from scipy.linalg import cholesky, LinAlgError

        print "Use numpy.linalg to compute eigen decomposition of covariance matrix"
        n_hor = self.state.shape[0]
        print "Matrix size: (%i x %i)"%(n_hor, n_hor)
        P = zeros((n_hor, n_hor))
        P_diag = zeros((n_hor, n_hor))
        iexp = {'e':1, 'g':2}[self.cortype]
        assert self.corlen >= 0, "ERROR - correlation length should be >= 0"
        # put stuff here to construct P for exponential or gaussian decay
        if self.corlen == 0 :
            P = eye(n_hor)
            P_diag = 1.*P
            lam = 1.
        else :
            # loop first index over latlon grid
            for p1 in self.state.itertuples():
                for p2 in self.state.itertuples():
                    dst = dist(p1.lon, p1.lat, p2.lon, p2.lat)
                    cor = exp(-(dst/self.corlen)**iexp)
                    P[p1.Index, p2.Index] = cor
                    P[p2.Index, p1.Index] = cor
            # Eigen decomposition of symmetric matrix
            #P0 = P[:]
            #try :
            #    P0 = cholesky(P0, lower=True)
            #except LinAlgError as err :
            #    print err.args[0]
            lam, P = eigh(P)
            lam = self.make_positive_semidef(lam)
            for i in xrange(n_hor):
                P_diag[i, i] = lam[i]

        self.n_hor = n_hor
        self.lam = lam
        self.P = P
        self.P_diag = P_diag

    def write(self, filename):
        ds = Dataset(filename, 'w')
        ds.nregions = 1
        ds.im = self.region.nlon
        ds.jm = self.region.nlat
        ds.lm = -1       # Not sure if this is used anywhere, so I prefer to put an invalid value, to be sure it'll crash if used.,
        ds.dx = self.region.dlon
        ds.dy = self.region.dlat
        ds.xref = ones(2)
        ds.yref = ones(2)
        ds.xbeg = self.region.lonmin
        ds.xend = self.region.lonmax
        ds.ybeg = self.region.latmin
        ds.yend = self.region.latmax
        ds.corlen = self.corlen
        ds.corchoice = self.cortype
        ds.createDimension('n_hor', self.n_hor)
        ds.createVariable('sqrt_lam', 'd', ('n_hor', ), zlib=True)
        ds.createVariable('lam', 'd', ('n_hor', ), zlib=True)
        ds.createVariable('P', 'd', ('n_hor', 'n_hor'), zlib=True)
        ds.variables['lam'][:] = self.lam
        ds.variables['sqrt_lam'][:] = self.lam**.5
        ds.variables['P'][:] = self.P.transpose() # numpy.eigv returns the vectors in columns, but the equivalent fotran subroutine returns them in rows. So store it in rows for consistency.

        # Also write B itself (first recalculate), for use in postprocessing
        print "Recalculating B matrix from eigenvectors/eigenvalues"
        ds.createVariable('B', 'd', ('n_hor', 'n_hor'), zlib=True)
        B = dot(self.P, self.P_diag)
        P = dot(B, self.P.transpose())
        ds.variables['B'][:] = P
        ds.close()

    def make_positive_semidef(self, lam):
        min_eigval = self.min_eigval+0.
        if self.min_eigval > 1.e-10:
            min_eigval = self.min_eigval*min((1., lam.max()))
        print "Maximum eigenvalue = %10.3e, minimum eigenvalue = %10.3e"%(lam.min(), lam.max())
        n_neg = sum(lam < min_eigval)
        lam[lam<min_eigval] = min_eigval
        if n_neg > 0 :
            print "Set %i eigenvalues to %15.11f"%(n_neg, min_eigval)
        return lam


def dist(lon1, lat1, lon2, lat2, ae=6.371e6):
    # Compute distance of two points on the globe
    # Based on TM5/misctools.F90/dist
    x1 = lon1*pi/180
    y1 = lat1*pi/180
    x2 = lon2*pi/180
    y2 = lat2*pi/180
    dy2 = (sin(0.5*(y2-y1)))**2
    dx2 = cos(y1)*cos(y2)*(sin(0.5*(x2-x1)))**2
    dd = 2*arcsin((dx2+dy2)**.5)
    ddg = dd*180/pi
    return (ddg*2*pi*0.001*ae)/360


def read_latlon(file_name):
    if not os.path.exists(file_name):
        raise RuntimeError("%s does not exist"%file_name)
    f = Dataset(file_name)
    P = f.variables['P'][:]
    D = f.variables['sqrt_lam'][:]
    f.close()
    # Thanks to Fortran's idiosyncracy, the indices of arrays in netcdf files are swapped
    # So P is actually transpose of the eigenvector matrix, and we need to transpose it back
    return transpose(P), D


def calc_temp_corr(corlen, dt, n):
    if corlen<1.e-20:
        P = eye(n)
        D = eye(n)
    else:
        dummy_X, dummy_Y = meshgrid(arange(n), arange(n))
        A = exp(-abs(dummy_X-dummy_Y)*dt/corlen)
        P,D = matrix_square_root(A)
        # Debug:
        # if abs((dot(dot(P, D**2), P.transpose())-A)) > 1.e-10 :
        #     raise SomeError
    return P, D


def matrix_square_root(B):
    # Given a real symmetric matrix B, calculate L such that LL^T = B
    # Actually, calculate L as the product of a unitary matrix and a diagonal matrix
    lam,P = linalg.eigh(B)
    # Sort eigenvalues and switch columns of P accordingly
    #sort_order = argsort(lam)
    sort_order = flipud(argsort(lam))
    lam = lam[sort_order]
    P = P[:, sort_order]
    D = diag(sqrt(lam))
    # Make sure that the elements in the top row of P are non-negative
    col_sign = where(P[0]<0.0, -1.0, 1.0)
    P = P*col_sign


   ## Select positive values :
   # ii, = where(lam>0)
   # lam = lam[ii]
   # P = P[:, ii]
   # D = diag(sqrt(lam))
    return P, D
    #lam[lam<0.0] = 0.0
    #D = diag(sqrt(lam))
    ## Make sure that the elements in the top row of P are non-negative
    #col_sign = where(P[0]<0.0, -1.0, 1.0)
    #P *= col_sign
    #return P, D

def g_to_gc(G_state, Temp_Lt, Hor_Lt, g, ipos, dummy):
    n_state = len(G_state)
    nt = shape(Temp_Lt)[0]
    nhor = shape(Hor_Lt)[0]
    #idx = zeros([nt*nt, 2])
    #k = 0
    #for i in xrange(nt):
    #    for j in xrange(nt):
    #        idx[k, :] = array([i, j])
    #        k +=1

    g_c = zeros([n_state])
    for i in tqdm(range(nt), desc='preconditioning gradient', leave=False):
        for j in xrange(nt):
            g_c[ipos+i*nhor:ipos+(i+1)*nhor] += dot(Temp_Lt[i,j]*Hor_Lt, G_state[ipos+j*nhor:ipos+(j+1)*nhor] * g[ipos+j*nhor:ipos+(j+1)*nhor])
        #print i, ipos+i*nhor, ipos+(i+1)*nhor, sum(g_c[ipos+i*nhor:ipos+(i+1)*nhor])
    return g_c

def xc_to_x(G_state, Temp_L, Hor_L, x_c, ipos, dummy):
    n_state = len(G_state)
    nt = shape(Temp_L)[0]
    nhor = shape(Hor_L)[0]

    x = zeros(n_state)
    for i in tqdm(range(nt), desc='xc_to_x', leave=True):
        for j in tqdm(range(nt), desc='step %i/%i'%(i, nt), leave=False):
            x[ipos+i*nhor:ipos+(i+1)*nhor] += G_state[ipos+i*nhor:ipos+(i+1)*nhor]* dot(Temp_L[i,j]*Hor_L, x_c[ipos+j*nhor:ipos+(j+1)*nhor])
    return x
