#!/usr/bin/env python

from numpy import isnan, dot, zeros_like, sqrt, inner, zeros, nan_to_num
from numpy.random import standard_normal
import os

class Optimizer(object):
    def __init__(self, rcf, control, model, minimizer):
        self.rcf = rcf              # Settings
        self.model = model          # Model interface instance (initialized!)
        self.control = control      # modelData structure, containing the optimization data
                                    # + aux variables for the transport model
        self.minimizer = minimizer  # minimizer object instance (initiated!)
        self.J = costFunction()

    def reset(self):
        self.minimizer.reset()

    def runPrior(self, label='apri'):
        state_preco = zeros(self.control.size)
        dy, err = self.computeDepartures(state_preco, label)
        self.J.bg, self.J.obs = self.computeCostFunction(state_preco, dy, err)
        self.save()

    def Var4D(self, label='apos'):
        self.reset()
        state_preco = zeros(self.control.size)
        status = 0
        step = 'apri'
        while status == 0 :
            state_preco, status = self.Var4D_step(state_preco, step)
            step = 'var4d'

        # Finishup ...
        dy, err = self.computeDepartures(state_preco, label)
        self.J.bg, self.J.obs = self.computeCostFunction(state_preco, dy, err)
        gradient_preco = self.ComputeGradient(state_preco, dy, err)
        self.minimizer.update(gradient_preco, self.J.tot)
        self.calcPosteriorUncertainties()
        self.save()

    def Var4D_step(self, state_preco, step='apri'):
        dy, err = self.computeDepartures(state_preco, step)
        self.J.bg, self.J.obs = self.computeCostFunction(state_preco, dy, err)
        gradient_preco = self.ComputeGradient(state_preco, dy, err)
        status = self.minimizer.calc_update(state_preco, gradient_preco, self.J.tot)
        state_preco = self.minimizer.readState()
        return state_preco, status

    def computeCostFunction(self, state_preco, dy, err):
        dstate = state_preco-self.control.read('state_prior_preco')
        J_bg = 0.5*dot(dstate, dstate)
        J_obs = 0.5*dot(dy/err, dy/err)
        return J_bg, J_obs

    def ComputeGradient(self, state_preco, dy, err):
        adjoint_state = self.computeObsGradient(dy/err**2)
        gradient_obs_preco = self.control.g_to_gc(adjoint_state)
        state_departures = state_preco-self.control.read('state_prior_preco')
        gradient_preco = gradient_obs_preco + state_departures
        return gradient_preco.values

    def computeDepartures(self, state_preco, step):
        state = self.control.xc_to_x(state_preco)
        self.control.store(state, 'state_%s'%step)
        emis = self.model.stateToStruct(state, self.control.categories)
        self.model.setupEmis(emis)
        departures = self.model.runForward(step)
        dy = departures.loc[:, 'mismatch']
        dye = departures.loc[:, 'err']
        minErr = self.rcf.get('observations.err.min')
        dye[isnan(dye)] = minErr
        dye[dye < minErr] = minErr
        return dy, dye

    def computeObsGradient(self, departures):
        adjoint_emis = self.model.runAdjoint(departures, self.control.categories)
        adjoint_state = self.model.stateToStruct_adj(adjoint_emis, self.control.categories)
        return adjoint_state

    def calcPosteriorUncertainties(self):
        converged_eigvals, converged_eigvecs = self.minimizer.read_eigsys()
        LE = zeros_like(converged_eigvecs)
        for ii in range(len(converged_eigvals)):
            # If use_eigen_file is true, for each eigenvector, check first whether the file eigenvector_modelspace_%03i.nn
            # exists or not. If it does, read the eigenvector in model space from that file. Else, calculate the
            # eigenvector in model space and store it in such a file.
            LE[:,ii] = self.control.xc_to_x(converged_eigvecs[:,ii], add_prior=False)
            self.control.store(LE[:, ii], 'eigenvec_%i'%ii)

        Mat2 = 1./converged_eigvals - 1.
        dapri = self.control.read('prior_uncertainty')
        dapos = nan_to_num(sqrt(dapri**2 + inner(LE**2, Mat2)))
        self.control.store(dapos, 'posterior_uncertainty')
        
    def readTrajectory(self):
        for istate, state_preco in tqdm(enumerate(self.minimizer.iter_states()), total=self.minimizer.iter):
            tqdm.write("Converting vector %i/%i to model space"%(istate, self.minimzer.iter))
            state = self.control.xc_to_x(state_preco)
            self.control.store(state, 'state_%i'%istate)
        
    def save(self, step=None):
        if step is None : 
            step = ''
        else :
            step += '.'
        path = self.rcf.get('path.output')
        if not os.path.exists(path): os.makedirs(path)
        self.rcf.write(os.path.join(path, 'lumia.%src'%step))
        self.model.save(path)
        self.control.write(os.path.join(path, 'control.%shdf'%step))
        self.minimizer.save(os.path.join(path, 'comm_file.%snc4'%step))
        #self.control.save(os.path.join(self.rcf.get('output.path'), 'optim_state.pickle'))
        #self.model.save(self.rcf.get('output.path'))

    def restart(self):
        # 1) Fill in self.control.vectors, as it was, from a savefile created in the prior run


        # 2) Restore the current state vector from the last saved location (i.e. comm_file.nc4)
        state_preco = self.minimizer.readState()

        # 3) Restore the cost function to what it was

        # 4) Re-do the last steps


class costFunction:
    def __init__(self):
        self.J_obs = None
        self.J_bg = 0

    def __getattr__(self, item):
        if item in ['J_obs', 'obs']:
            return self.J_obs
        if item in ['bg', 'J_bg','b']:
            return self.J_bg
        elif item in ['J', 'J_tot', 'tot']:
            try :
                return self.J_bg+self.J_obs
            except TypeError :
                raise RuntimeError("J_tot cannot be computed because J_obs hasn't been evaluated")
        else :
            raise AttributeError("CostFunction attributes can only be J, tot, J_tot, J_bg or J_obs")

    def __setattr__(self, key, value):
        # Modify __dict__ directly to avoid infinite recursion loops
        if key in ['obs', 'J_obs']:
            self.__dict__['J_obs'] = value
        elif key in ['bg', 'J_bg']:
            self.__dict__['J_bg'] = value
        else :
            raise AttributeError("Attribute %s not permitted"%key)


class Categories:
    def __init__(self, rcf=None):
        self.list = []
        if rcf is not None :
            self.setup(rcf)

    def add(self, name):
        if not name in self.list :
            setattr(self, name, Category(name))
            self.list.append(name)
        else :
            raise RuntimeError("Category %s already exists!"%name)

    def __getitem__(self, item):
        if item in self.list :
            return getattr(self, item)
        else :
            raise KeyError("Category %s not initialized yet!"%item)

    def __iter__(self):
        for item in self.list :
            yield getattr(self, item)

    def setup(self, rcf):
        catlist = rcf.get('emissions.categories')
        for cat in catlist :
            self.add(cat)
            self[cat].optimize = rcf.get('emissions.%s.optimize'%cat, totype=bool)
            self[cat].is_ocean = rcf.get('emissions.%s.is_ocean'%cat, totype=bool, default=False)
            if self[cat].optimize :
                self[cat].uncertainty = rcf.get('emissions.%s.error'%cat)
                self[cat].uncertainty_type = rcf.get('emissions.%s.error_type'%cat, default='tot')
                self[cat].min_uncertainty = rcf.get('emissions.%s.error_min'%cat, default=0)
                self[cat].horizontal_correlation = rcf.get('emissions.%s.corr'%cat)
                self[cat].temporal_correlation = rcf.get('emissions.%s.tcorr'%cat)
                self[cat].optimization_interval = rcf.get('optimization.interval')

    def __len__(self):
        return len(self.list)


class Category:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other
