#!/usr/bin/env python

from lumia.optimizer import Categories
from lumia.Tools import Region
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from numpy import *
import os
from netCDF4 import Dataset
from lumia.comm import communicator
from lumia.tools.maptools import refine_map
from pandas import read_hdf
from tqdm.autonotebook import tqdm


class Preprocessor(communicator):
    def __init__(self, rcf):
        communicator.__init__(self, 0)
        self.rcf = rcf
        self.categories = Categories(rcf)
        self.region = Region(rcf)
        self.tmin = datetime(*self.rcf.get('time.start'))
        self.tmax = datetime(*self.rcf.get('time.end'))

    def createEmissionsFile(self, filename):
        em = self.createEmissionsStructure()
        writeEmis(filename, em)

    def createEmptyStructure(self):
        em = {}
        for cat in self.categories :
            interval_emis = self.rcf.get('emissions.%s.interval'%cat.name)
            dt = {
                'y':relativedelta(years=1),
                'm':relativedelta(months=1),
                'd':timedelta(1),
                '3h':timedelta(hours=3)
            }[interval_emis]
            tt = self.tmin
            time_start = []
            time_end = []
            while tt < self.tmax :
                time_start.append(tt)
                time_end.append(tt+dt)
                tt += dt
            nt = len(time_start)

            em[cat.name] = {
                'emis':zeros((nt, self.region.nlat, self.region.nlon)),
                'time_interval':{
                    'time_start':array(time_start),
                    'time_end':array(time_end)
                }
            }

            em['cat_list'] = []
        return em

    def createEmissionsStructure(self):
        em = self.createEmptyStructure()
        for cat in self.categories :
            interval = self.rcf.get('emissions.%s.interval'%cat.name)
            em[cat.name]['emis'] = self.readEmis(cat.name, em[cat.name], interval)
            em['cat_list'].append(cat.name)
        return em

    def readEmis2(self, catname, emstruct, interval):
        nt = emstruct['emis'].shape[0]
        for itime in xrange(nt):
            # 1) Check if pre-processed files exist, if not, create it
            pp = PpFile(
                catname,
                self.rcf.get('emissions.pp.path'),
                interval,
                self.region,
                emstruct['time_interval']['time_start'][itime],
                emstruct['time_interval']['time_end'][itime],
                self.rcf.get('emissions.%s.class'%catname)
            )
            if not pp.check():
                pp.make(self.rcf)
            else :
                emstruct['emis'][itime, :, :] = pp.read()
        return emstruct['emis']

    def readEmis(self, catname, emstruct, interval):
        """
        Reads emissions in a pre-processed pandas DataFrame format.
        - The dataFrame needs to have 4 columns (at least): lat, lon, time and the flux to be read
          (several fluxes can be found in one dataframe, sharing the same coordinates).
        - the dataFrames are stored in monthly HDF files
        - the files are assumed to be in umolCO2/m2/s
        """

        tstart = emstruct['time_interval']['time_start']
        tend = emstruct['time_interval']['time_end']
        months = array([datetime(x.year, x.month, 1) for x in tstart])
        empath = os.path.join(self.rcf.get('emissions.df.path'), self.region.name, interval)
        emprefix = self.rcf.get('emissions.%s.prefix'%catname)
        field = self.rcf.get('emissions.%s.class'%catname)

        for month in tqdm(unique(months), leave=False, desc=catname):
            select_t = months == month 
            tmin = amin(tstart[select_t])
            tmax = amax(tend[select_t])

            df = read_hdf(os.path.join(empath, '%s%s.hdf'%(emprefix, month.strftime('%Y%m'))))
            df = df.loc[(df.time >= tmin) & (df.time < tmax)]
            flux = df.loc[:, field].values
            lats = df.loc[:, 'lat'].values
            lons = df.loc[:, 'lon'].values
            # Just in case, check that the dimensions are compatible.
            # Here would be the place to introduce a mask to reduce the domain size if needed]
            if not array_equal(self.region.lats, unique(lats)):
                self.error("Latitudes of emission files does not match requirements")
            if not array_equal(self.region.lons, unique(lons)):
                self.error("Longitudes of emission files does not match requirements")

            emstruct['emis'][select_t, :, :] = flux.reshape(-1, self.region.nlat, self.region.nlon)

        return emstruct['emis']


class TM5_Emis(Preprocessor):
    def __init__(self, rcf, emfile, match_cat={'fire':'fires'}):
        Preprocessor.__init__(self, rcf)
        self.emfile = emfile                      # Name of the file from which to read the emissions
        #self.dailyCyclePath = dailyCyclePath      # Name of the path from which to read the daily cycle coefficients
        self.match_cat = match_cat                # dictionary of category name correspondances
        #self.preloaded = []
        self.TM5_data = self.loadEmissions()                      # Load the raw emissions (fill in self.TM5_data)

#    def preload(self, categ=None, tmin=None, tmax=None):
#        if not categ in self.preloaded: self.applyDailyCycles(categ, tmin, tmax)
#        self.preloaded.append(categ)
#
#    def applyDailyCycles(self, categ=None, tmin=None, tmax=None, reg=None):
#        """
#        This applies the daily cycle using the daily cycle files in the specified folder
#        The fluxes are returned in kgCO2/m2/sec
#        This is the main loop, the actual daily cycle is computed in the self.applyDailyCycle method, for each time-step/region
#        """
#        if reg is None :
#            regions = [x for x in self.TM5_data.keys() if not x in ['time']]
#        else :
#            regions = [reg]
#        for reg in regions :
#            if categ is None :
#                cats = self.TM5_data[reg]['emis'].keys()
#            else :
#                cats = [categ]
#            for cat in cats :
#                times, emh = self.applyDailyCycle(cat, self.TM5_data[reg]['region'], tmin, tmax)
#                self.TM5_data[reg]['emis_hourly'] = {
#                    cat     : emh,
#                    'times' : times
#                }
#
#    def applyDailyCycle(self, category, reg, tmin=None, tmax=None):
#        """
#        This applies the daily cycle using the daily cycle files in the specified folder, for a given category, region and time interval
#        The fluxes are returned in kgCO2/m2/sec
#        """
#        if tmin is None : tmin = self.TM5_data['time']['start'].min()
#        if tmax is None : tmax = self.TM5_data['time']['end'].max()
#            
#        ntstep_hourly = 8*(tmax-tmin).days
#        
#        emcat = zeros((ntstep_hourly, self.TM5_data[reg.name]['region'].nlat, self.TM5_data[reg.name]['region'].nlon))
#        
#        times_start_hourly = array([tmin + timedelta(hours=3)*x for x in range(ntstep_hourly)])
#        month_start_hourly = array([datetime(x.year, x.month, 1) for x in times_start_hourly])
#        times_start = self.TM5_data['time']['start'][(self.TM5_data['time']['start'] >= tmin)*(self.TM5_data['time']['end'] <= tmax)]
#
#        # Fill in the array with the base fluxes
#        # Fluxes are in kgCO2
#        for time in times_start:
#            emcat[month_start_hourly == time, :, :] = self.TM5_data[reg.name]['emis'][self.match_cat.get(category, category)][self.TM5_data['time']['start'] == time, :, :]
#
#        # Apply the daily cycles
#        days = array([datetime(x.year, x.month, x.day) for x in times_start_hourly])
#        for day in unique(days) :
#            try :
#                ds = Dataset(os.path.join(self.dailyCyclePath, day.strftime('%Y/%m/CO2_dailyCycle_%Y%m%d.nc4')), 'r')
#            except IOError :
#                self.IOError(os.path.join(self.dailyCyclePath, day.strftime('%Y/%m/CO2_dailyCycle_%Y%m%d.nc4')))
#            anomaly = ds[reg.name][self.match_cat.get(category, category)]['emission_anomaly'][:]
#            emcat[days == day, :, :] += anomaly
#            ds.close()
#
#        # convert to kg/m2/s because it's more convenient for the regridding
#        emcat /= reg.area/timedelta(hours=3).total_seconds()
#
#        # Build the time array
#        from lumia.tools import tinterv
#        times_out = [tinterv(tt, tt+timedelta(hours=3)) for tt in times_start_hourly]
#        return times_out, emcat

    def loadEmissions_old(self, tracer='CO2'):
        """
        This reads the emissions from a TM5 emission file (emission.nc4) into a dictionary
        The emissions are in kgCO2/gridbox/time-step (i.e. kgCO2)
        """
        
        from lumia.Tools import region
        ds = Dataset(self.emfile)
        TM5_data = {}
        for reg in ds.groups :
            TM5_data[reg] = {
                'region': region(
                    lat0=ds[reg].latmin,
                    lat1=ds[reg].latmax,
                    lon0=ds[reg].lonmin,
                    lon1=ds[reg].lonmax,
                    nlat=ds[reg].dimensions['latitude'].size,
                    nlon=ds[reg].dimensions['longitude'].size,
                    name=reg
                ),
            }
            TM5_data[reg]['emis'] = {}
            for cat in ds[reg][tracer].groups:
                TM5_data[reg]['emis'][self.match_cat.get(cat, cat)] = ds[reg][tracer][cat]['emission'][:]

            # Copy the times only once as they are the same for all categories
            # I know they may be different, but for now, we assume they aren't
            TM5_data['time'] = {
                'start' : array([datetime(*x) for x in ds[reg][tracer][cat]['time_start']]),
                'end': array([datetime(*x) for x in ds[reg][tracer][cat]['time_end']])
            }
        ds.close()
        return TM5_data
    
    def loadEmissions(self, tracer='CO2'):
        import pickle
        from lumia.Tools import region
        
        # Load data in memory
        with open(self.emfile, 'r') as fid :
            TM5_data = pickle.load(fid)
            
        # Reshape the emissions in a more convenient way
        catlist = TM5_data['glb600x400'].keys()
        TM5_data['glb600x400']['emis'] = {}
        for cat in catlist:
            emis = TM5_data['glb600x400'][cat]['emis'][0]
            for em in TM5_data['glb600x400'][cat]['emis'][1:]:
                emis = append(emis, em, axis=0)
            TM5_data['glb600x400']['emis'][self.match_cat.get(cat, cat)] = emis
            # Save tmin/tmax for later
            tmin = TM5_data['glb600x400'][cat]['time'][0]
            tmax = TM5_data['glb600x400'][cat]['time'][-1]
            del TM5_data['glb600x400'][cat]
            
        # Create the region definition
        TM5_data['glb600x400']['region'] = region(lat0=-90,lat1=90,lon0=-180,lon1=180,dlat=4,dlon=6,name='glb600x400')
        
        # Create the time definition
        tmin = datetime(tmin.year, tmin.month, 1)
        tmax = datetime(tmax.year, tmax.month, 1)+relativedelta(months=1)
        times = [tmin]
        tt = tmin
        while tt <= tmax:
            tt += relativedelta(hours=3)
            times.append(tt)
        TM5_data['time'] = {
            'start' : array(times)[:-1],
            'end':array(times)[1:]
        }
        
        # The emissions are at this stage in kgCO2/sec. We want kgCO2/m2/sec:
        for cat in catlist :
            TM5_data['glb600x400']['emis'][self.match_cat.get(cat, cat)] /= TM5_data['glb600x400']['region'].area
        
        return TM5_data
    
#    def rebin_time(self, flux, emstruct, catname):
#        emis_fine = zeros((emstruct['emis'].shape[0], flux['glb600x400']['emis'][catname].shape[1], flux['glb600x400']['emis'][catname].shape[2]))
#        for it, (ti, tf) in enumerate(zip(emstruct['time_interval']['time_start'], emstruct['time_interval']['time_end'])):
#            it_flux = flux['time']['start'].tolist().index(datetime(ti.year, ti.month, 1))
#            dt = (flux['time']['end'][it_flux]-flux['time']['start'][it_flux]).total_seconds()
#            emis_fine[it, :, :] = flux['glb600x400']['emis'][catname][it_flux, :, :]/dt/flux['glb600x400']['region'].area
#        return emis_fine

    def readEmis(self, catname, emstruct, interval):
        # Apply the daily cycles for that category
        # self.preload(catname)
        
        #flux_m = self.loadEmissions()                          # Flux in kgCO2/gridbox/tstep
        #flux_3h = self.rebin_time(flux_m, emstruct, catname)  # Flux in kgCO2/m2/sec
        flux_3h = self.TM5_data['glb600x400']['emis'][catname]
        reg_tm5 = self.TM5_data['glb600x400']['region']
        
        regridded = []
        for itime in xrange(flux_3h.shape[0]):
            regridded.append(refine_map(flux_3h[itime, :, :], reg_tm5, self.region))
        # At this stage, emissions are in kgCO2/m2/sec. We want umol/m2/s:
        return array(regridded)*1.e9/44

        # Extract the data at each time interval
        #region_tm5 = 'glb600x400'
        #nt = emstruct['emis'].shape[0]

        #regridded = []
#
#        for itime in xrange(nt):
#            # The following will crash if the requested flux and the TM5 flux are not on the same temporal grid
#            start = emstruct['time_interval']['time_start'][itime]
#            end = emstruct['time_interval']['time_end'][itime]
#            times = array(self.TM5_data[region_tm5]['emis_hourly']['times'])
#            times_selected = times[(times >= start)*(times <= end)]
#            if len(times_selected) != 1 : import pdb; pdb.set_trace()
#            time = times_selected[0]
#
#            regridded.append(refine_map(
#                squeeze(self.TM5_data[region_tm5]['emis_hourly'][catname][times == time, :, :])*time.dt.total_seconds(),
#                self.TM5_data[region_tm5]['region'],
#                self.region
#            ))
            
        # At this stage, emissions are in kgCO2/m2/sec. We want umol/m2/s:
        #return array(regridded)*1.e9/44

        # reconvert to kgCO2
        #regridded = [x*self.region.area for x in regridded]
        
        # Finally, we want umol/m2/s
        #return array(regridded)*1.e9/44/region.area/(tmax-tmin).total_seconds()


def writeEmis(filename, emis):
    """
    Write the emissions as needed by the transport model
    """
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))
    with Dataset(filename, 'w') as ds :
        ds.createDimension('time_components', 6)
        for cat in emis.keys():
                gr = ds.createGroup(cat)
                gr.createDimension('nt', emis[cat]['emis'].shape[0])
                gr.createDimension('nlat', emis[cat]['emis'].shape[1])
                gr.createDimension('nlon', emis[cat]['emis'].shape[2])
                gr.createVariable('emis', 'd', ('nt', 'nlat', 'nlon'))
                gr['emis'][:] = emis[cat]['emis']
                gr.createVariable('times_start', 'i', ('nt', 'time_components'))
                gr['times_start'][:] = array([x.timetuple()[:6] for x in emis[cat]['time_interval']['time_start']])
                gr.createVariable('times_end', 'i', ('nt', 'time_components'))
                gr['times_end'][:] = array([x.timetuple()[:6] for x in emis[cat]['time_interval']['time_end']])


class PpFile(object):
    def __init__(self, catname, path, interval, region, tmin, tmax, classname):
        self.filePath = os.path.join(path, region.name, interval)
        fmt = '%Y%m%d%H'
        filename = '%s.%s-%s.nc'%(catname, tmin.strftime(fmt), tmax.strftime(fmt))
        self.fileName = os.path.join(self.filePath, filename)
        self.field = 'flux.%s'%classname

        # pre-processed files are in units of kgCO2, we want umol/m2/s
        self.factor = 1.e9/44/region.area/(tmax-tmin).total_seconds()

    def check(self):
        valid = os.path.exists(self.fileName)
        if valid :
            ds = Dataset(self.fileName, 'r')
            valid *= self.field in ds.variables
            ds.close()
        return valid

    def make(self, rcf):
        print "checking %s:%s"%(self.fileName, self.field)
        raise NotImplementedError

    def read(self):
        with Dataset(self.fileName, 'r') as ds :
            emis = ds[self.field][:]*self.factor
        return emis
