#!/usr/bin/env python
from numpy import unique, isnan, array, zeros_like, zeros, shape
from copy import deepcopy
import os, tempfile, subprocess, pickle
from netCDF4 import Dataset
from dateutil.relativedelta import relativedelta
from datetime import timedelta, datetime
from optimData import coarsenTime
from pandas import isnull

class Interface(object):
    def __init__(self, rcf, params, observations):
        self.rcf = rcf
        self.region = params.region
        self.db = observations
        self.emis = None
        self.stateInfo = params.vectors.loc[:, ('category', 'time_start', 'time_end')]
        if 'prior_emis' in params.model_data:
            self.setupEmis(params.model_data['prior_emis'])

    def stateToStruct(self, state, catInfo, refinetime=True):
        emis = deepcopy(self.emis)
        for cat in [x for x in catInfo if x.optimize]:
            state_cat = state[self.stateInfo.category == cat.name]
            stateInfo_cat = self.stateInfo.loc[self.stateInfo.category == cat.name]
            nst = stateInfo_cat.shape[0]
            time_start = unique([stateInfo_cat.loc[x, 'time_start'].to_pydatetime() for x in xrange(nst)])
            time_end = unique([stateInfo_cat.loc[x, 'time_end'].to_pydatetime() for x in xrange(nst)])
            nt = len(time_start)
            emcat = state_cat.values.reshape(nt, self.region.nlat, self.region.nlon)
            emcat = {
                'emis':emcat,
                'time_interval':{
                    'time_start': time_start,
                    'time_end' :time_end
                }
            }
            if refinetime: # We may not want to do this sometimes ...
                emis[cat.name]['emis'] = refineTime(emcat, self.prior_emis[cat.name])
            else :
                import pdb; pdb.set_trace()
                raise RuntimeError("This is clearly a bug")
        return emis

    def stateToStruct_adj(self, adjoint_emis, catInfo):
        """
        Generates an adjoint state vector from the adjoint emission structure returned by the adjoint
        model. This requires in particular applying the adjoint of the disaggregation function used
        to generate high resolution emissions from low resolution state vector
        :param adjoint_emis: the adjoint emission structure as returned by the runAdjoint method
        :param catInfo: the list of categories
        :return: the adjoint state vector that corresponds do dJ_obs
        """

        dt = {'y':relativedelta(years=1), 'm':relativedelta(months=1), 'd':timedelta(1)}
        adjState = zeros(self.stateInfo.shape[0])
        catIndex = self.stateInfo.loc[:, 'category'].values

        for cat in [x for x in catInfo if x.optimize]:
            #1) Adjoint of refineTime (#TODO: take this out of stateToStruct_adj)
            adjCat = refineTime_adj(
                deepcopy(adjoint_emis[cat.name]),
                self.prior_emis[cat.name],
                dt[cat.optimization_interval]
            )

            # 2) Coarsen:
            adjCat = coarsenTime(adjCat, cat.optimization_interval)

            # 3) Fill in the state:
            stateCat = adjCat['emis'].reshape(-1)
            adjState[catIndex == cat] = stateCat

        return adjState

    def setupEmis(self, emstruct):
        self.emis = emstruct
        self.prior_emis = deepcopy(self.emis)

    def writeEmis(self, filename=None, emis=None):
        """
        Write the emissions as needed by the transport model
        """
        if filename is None : filename = self.rcf.get('emissions.filename')
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))
        if emis is None : emis = self.emis
        with Dataset(filename, 'w') as ds :
            ds.createDimension('time_components', 6)
            for cat in emis['cat_list']:
                gr = ds.createGroup(cat)
                gr.createDimension('nt', emis[cat]['emis'].shape[0])
                gr.createDimension('nlat', emis[cat]['emis'].shape[1])
                gr.createDimension('nlon', emis[cat]['emis'].shape[2])
                gr.createVariable('emis', 'd', ('nt', 'nlat', 'nlon'))
                gr['emis'][:] = emis[cat]['emis']
                gr.createVariable('times_start', 'i', ('nt', 'time_components'))
                gr['times_start'][:] = array([x.timetuple()[:6] for x in emis[cat]['time_interval']['time_start']])
                gr.createVariable('times_end', 'i', ('nt', 'time_components'))
                gr['times_end'][:] = array([x.timetuple()[:6] for x in emis[cat]['time_interval']['time_end']])
        return filename

    def runForward(self, step):
        #TODO: move the parallelization to the transport model itself?
        executable = self.rcf.get("model.transport.exec")
        select = self.db.observations.footprint_exists.values

        nchunks = self.rcf.get('model.transport.split', default=1)
        nobs = len(self.db.observations.loc[select])
        chunk_size = nobs/nchunks
        remain = nobs%nchunks
        if remain != 0 : chunk_size += 1

        # Write emissions (common to all chunks)
        emf = self.writeEmis(os.path.join(self.rcf.get('path.run'), 'emissions.%s.nc4'%step))

        # Run the individual transport model chunks
        pids = []
        files = []
        for ichunk in range(nchunks):
            # Write observation file
            dbfid, dbf = tempfile.mkstemp(
                dir=self.rcf.get('path.run'),
                prefix='forward.',
                suffix='.db'
            )

            # The previous function actually creates a file. We need to remove it or we end up with too many open files
            os.fdopen(dbfid).close()

            # We actually just need these two fields:
            db = self.db.observations.loc[select].iloc[ichunk*chunk_size:(ichunk+1)*chunk_size]
            db.loc[:, ['footprint', 'time']].to_pickle(dbf)

            # Launch the transport subprocesses
            print ' '.join([x for x in ['python', executable, '--forward', '--db', dbf, '--emis', emf]])
            pids.append(subprocess.Popen(
                ['python', executable, '--forward', '--db', dbf, '--emis', emf],
                close_fds=True
            ))
            files.append(dbf)

            # Let the subprocesses finish
        sigterm = [x.wait() for x in pids]

        # Retrieve the data
        for dbf in files :
            with open(dbf, 'r') as fid :
                db = pickle.load(fid)
            try :
                iid = db.id[-isnan(db.id)]
                self.db.observations.loc[iid, 'foreground'] = db.loc[iid, 'foreground']
                self.db.observations.loc[iid, 'totals'] = db.loc[iid, 'totals']
                self.db.observations.loc[iid, 'model'] = db.loc[iid, 'model']
            except IOError:
                import pdb; pdb.set_trace()
            except AttributeError:
                print dbf
                import shutil
                shutil.copy(dbf, os.environ['HOME'])
                print shape(db)
                raise AttributeError

            # Cleanup the temporary files:
            os.remove(dbf)

        self.db.observations.loc[:, 'mismatch'] = \
            self.db.observations.loc[:,'background'] + \
            self.db.observations.loc[:,'foreground'] - \
            self.db.observations.loc[:,'obs']

        self.db.observations.loc[:, step] = self.db.observations.loc[:, 'background']+self.db.observations.loc[:, 'foreground']

        selection = ~isnull(self.db.observations.mismatch)
        return self.db.observations.loc[selection, ('mismatch', 'err')]

    def runAdjoint(self, departures, categories):
        from tools import dict2n4, n4todict

        #TODO: move the parallelization to the transport model itself?
        executable = self.rcf.get("model.transport.exec")

        # Create the adjoint structure
        cat_optim = [x for x in categories if x.optimize]
        adj = {'cat_list':[x.name for x in cat_optim]}
        for cat in cat_optim:
            adj[cat.name] = {
                'emis':zeros_like(self.emis[cat.name]['emis']),
                'time_interval':self.emis[cat.name]['time_interval']
            }

        # Create the departures database
        minerr = self.rcf.get('observations.err.min')
        dp = self.db.observations.loc[:, ['footprint', 'time']]
        dp.loc[:, 'dy'] = departures #self.db.observations.loc[:, 'mismatch']/err**2

        # Compute splitting variables
        nchunks = self.rcf.get('model.transport.split', default=1)
        nobs = len(self.db.observations)
        chunk_size = nobs/nchunks
        remain = nobs%nchunks
        if remain != 0 : chunk_size += 1

        # Run the individual transport model chunks
        pids = []
        dpfiles, adjfiles = [], []
        for ichunk in range(nchunks):

            # Write the departures file :
            dpfid, dpf = tempfile.mkstemp(
                dir=self.rcf.get('path.run'),
                prefix='departures.',
                suffix='.db'
            )
            dp.iloc[ichunk*chunk_size:(ichunk+1)*chunk_size].to_pickle(dpf)
            dpfiles.append(dpf)

            # Open and close the file descriptor to avoid "too many open files" error
            os.fdopen(dpfid).close()

            # Write adjoint
            adjf = dpf.replace('departures', 'adjoint')
            dict2n4(adj, adjf)
            adjfiles.append(adjf)

            cmd = ['python', executable, '--adjoint', '--db', dpf, '--emis', adjf]
            pids.append(subprocess.Popen(cmd))

        # Wait for the subprocesses to finish
        sigterm = [pid.wait() for pid in pids]

        # Collect the results :
        for adjf in adjfiles :
            adjx = n4todict(adjf)
            for cat in cat_optim :
                adj[cat.name]['emis'] += adjx[cat.name]['emis']

        # Remove the temporary files :
        for file in adjfiles+dpfiles : os.remove(file)

        return adj

    def save(self, path):
        self.db.write(os.path.join(path, 'observations.db'))


def refineTime(coarseEmis, fineEmis):
    """
    Restore the high resolution temporal structure of the fluxes, from that of a previous time step.
    With:
        - x the monthly total flux at one grid point, 
	- f(t) the flux at time step t,
	- nt the number of time steps in a month
	- f0(t) the previous (or prior, it does not matter) flux at time step t:
    f(t) = x/nt + f0 - x0/nt

    x is provided in coarseEmis,
    f0 and x0 are deduced from fineEmis
    nt is computed
    """
    f = deepcopy(fineEmis)
    nt_optim = len(coarseEmis['time_interval']['time_start'])
    tstart_emis = fineEmis['time_interval']['time_start']
    for i_time in xrange(nt_optim):
        tmin = coarseEmis['time_interval']['time_start'][i_time]
        tmax = coarseEmis['time_interval']['time_end'][i_time]
        select = (tstart_emis >= tmin)*(tstart_emis < tmax)
        nt = sum(select)+0.   # Make sure we have a real
        x1 = coarseEmis['emis'][i_time, :, :]
        f0 = fineEmis['emis'][select, :, :]
        x0 = f0.sum(0)
        f['emis'][select, :, :] = x1/nt + f0 - x0/nt

    return f['emis']


def refineTime_adj(adjEmis, fineEmis, dt_optim):
    intervals_adjEmis = adjEmis['time_interval']['time_start']
    if dt_optim == relativedelta(years=1):
        intervals_optim = array([datetime(x.year, 1, 1) for x in intervals_adjEmis])
    elif dt_optim == relativedelta(months=1):
        intervals_optim = array([datetime(x.year, x.month, 1) for x in intervals_adjEmis])
    elif dt_optim == timedelta(1) :
        intervals_optim = array([datetime(x.year, x.month, x.day) for x in intervals_adjEmis])
    else :
        raise NotImplementedError
    for i_time, time in enumerate(unique(intervals_optim)) :
        select = intervals_optim == time
        nt = sum(select)+0.  # Make sure we have a real
        adjEmis['emis'][select, :, :] /= nt
    return adjEmis

