#!/usr/bin/env python

"""
Example script
"""

import lumia
from numpy import *
from lumia.tools.emissions import Preprocessor

rcf = lumia.rcf('example.rc')

# Set a "tag" to the simulation (i.e. give it a name) by modifying (or creating) the "tag" rc key
rcf.setkey('tag', 'myinv')

# Read a pre-existing observation database:
obs = lumia.obsdb()
obs.read_hdf(rcf.get('file.obsdb'))

# The observation has two main objects:
# obs.observations is a pandas dataframe containing the individual observations, and variables in the observation space
# The number of variables is free (only obs, err, background and footprint are really mandatory for the inversion). The
# user can for instance import concentrations from a previous estimate or from another model:
# obs.observations.loc[:, 'model2'] = model2conc
#
# obs.sites contains the list of observation sites, and some of their characteristics (sitename, sampling height, code, etc.)
# Again, the number of variables is free. For instance one can tag some sites as part of subsets for sensitivity inversions
#
# By default, the obs.observations dataframe contains a "site" column, which refers to the index of "obs.sites"

# Setup the obs uncertainties :
obserrmin = rcf.get('observations.err.min')
errmin = rcf.get('observations.errtot.min')

obs.observations.loc[obs.observations.err_obs < obserrmin, 'err_obs'] = obserrmin
errtot = sqrt(obs.observations.err_obs**2+obs.observations.err_mod**2+obs.observations.err_profile_bg**2)
errtot[errtot<errmin] = errmin
obs.observations.loc[:, 'err'] = errtot

# Exclude some observation sites :
obs.excludeSites(rcf.get('sites.excluded'))

# Check that observation footprints are present:
obs.checkFootprints(rcf.get('footprints.path'))

# Load emission files (i.e. generate an emission dictionary following the structure:
# em = {'cat_list':['cat1','cat2','cat3'],   # list of emission categories
#       'cat1':{
#               'emis':data                  # with data a 3D numpy array (ntime, nlat, nlon)
#               'time_interval':{
#                   'time_start':time_start  # array of datetime.datetime objects marking the start of the flux period
#                   'time_end'  :time_end    # array of datetime.datetime objects marking the end of the flux period
#        },
#        'cat2': ...
#}
# Emissions are assumed to be in umol/m2/s
pp = Preprocessor(rcf)
emstruct = pp.createEmissionsStructure()

# Initialize the optimData object with the emissions we have just imported
emis = lumia.optimData(rcf, prior=emstruct)
emis.setupUncertainties()

#######################################################
# Setup the transport, optimizer and minimizer objects
model = lumia.transport(rcf, emis, obs)
minim = lumia.Minimizer(rcf, emis.size)
opt = lumia.Optimizer(rcf, emis, model, minim)

#######################################################
# Run the inversion
opt.Var4D(label=rcf.get('tag'))
