module covar_latlon
    use linalg_interface
    use file_netcdf
    use netcdf
    use misctools

    implicit none

    public :: calc_latlon_covariance, read_vec2ll, min_eigval, save_latlon_covariance
    private

    integer         :: n_hor
    real            :: min_eigval

    ! Vec2ll
    ! Mapping of state vector space on lat-lon grid
    ! No need to load all the tm5 stuffs. Only lon and lat are needed here.
    type Tvec2ll
        real :: lon, lat
    end type Tvec2ll

    type(Tvec2ll), dimension(:), allocatable :: vec2ll

    real, dimension(:,:), allocatable :: P, B, P_diag
    real, dimension(:), allocatable   :: lam, lam_sqrt

contains
    subroutine read_vec2ll(fname)
        character(len=*), intent(in) :: fname
        integer :: ncf, dimid, status
        real, dimension(:), allocatable :: tmp
        ncf = nc_open(trim(fname), 'r')
        status = nf90_inq_dimid(ncf, 'n_hor', dimid)
        status = nf90_inquire_dimension(ncf, dimid, len=n_hor)
        allocate(vec2ll(n_hor), tmp(n_hor))
        tmp = nc_read_var(ncf, 'vec2ll_lon')
        vec2ll(:)%lon = tmp
        tmp = nc_read_var(ncf, 'vec2ll_lat')
        vec2ll(:)%lat = tmp
        deallocate(tmp)
        call nc_close(ncf)
    end subroutine read_vec2ll

    subroutine calc_latlon_covariance(corlen, choice)

        ! --- in/out ----------------------------------------------
        real, intent(in)             :: corlen
        character(len=1), intent(in) :: choice

        ! --- local -----------------------------------------------
        logical                            :: diagonalized
        integer                            :: i_hor1, i_hor2
        real                               :: lon1, lat1, lon2, lat2, dst, cor
        integer                            :: i, iexp, mat_size

        ! --- begin -----------------------------------------------
        allocate(P(n_hor,n_hor), P_diag(n_hor,n_hor))
        allocate(lam(n_hor), lam_sqrt(n_hor))

        write(*,'("calc_latlon_covariance: corlen = ",f7.1," ; choice = ",a1)') corlen, choice
        write(*,*) "calculating covariance matrix (may take a while)"

        ! Function for correlations
        select case ( choice )
        case ( 'e' )
            iexp = 1
        case ( 'g' )
            iexp = 2
        case default
            write (*,'("ERROR - invalid choice for spatial correlation: ",a1)') choice
            call exit(1)
        end select

        if (corlen < 0) then
            write (*,'("ERROR - correlation length should be >= 0")')
            call exit(1)
        endif

        diagonalized = .false. ! P has not been diagonalized yet

        ! put stuff here to construct P for exponential or gaussian decay
        if ( corlen == 0.0 ) then
            P = 0.
            P_diag = 0.
            do i_hor1 = 1, n_hor
                P(i_hor1,i_hor1) = 1.
                P_diag(i_hor1,i_hor1) = 1.
            enddo
            lam = 1.
            diagonalized = .true.
        else
            ! loop first index over latlon grid
            do i_hor1 = 1, n_hor
                print*, i_hor1, '/', n_hor
                lon1 = vec2ll(i_hor1)%lon
                lat1 = vec2ll(i_hor1)%lat
                ! loop second index over latlon grid
                do i_hor2 = i_hor1, n_hor
                    lon2 = vec2ll(i_hor2)%lon
                    lat2 = vec2ll(i_hor2)%lat
                    call dist( lon1, lat1, lon2, lat2, dst )
                    cor = exp( -(dst/corlen)**iexp )
                    P(i_hor1,i_hor2) = cor
                    P(i_hor2,i_hor1) = cor
                enddo
            enddo
        endif

        if (.not. diagonalized) then
            ! Eigen decomposition of symmetric matrix
            mat_size = size(P, 1)
            print*, 'decompose'
            call eigvals(mat_size, P, lam, .true.)
            print*, 'make positive semidef'
            call make_positive_semidef(lam)

            lam_sqrt = sqrt( lam )   ! Return square root
            P_diag = 0.0
            do i=1,n_hor
                print*, i, '/', n_hor
                P_diag(i,i) = lam(i)
            end do
            diagonalized = .true.
        end if
    end subroutine calc_latlon_covariance

    subroutine save_latlon_covariance(fname)
        character(len=*), intent(in) :: fname
        integer         :: nc_id

        ! write to file for later use:
        ! turn on compression for the variables
        nc_variables_deflate = .true.
        write(*,'("Writing covariance file to disk: ", a)') trim(fname)

        ! If the folder with the Bh files does not exist, create it
        call check_dir(fname)

        nc_id = nc_open(fname, 'c')
        call nc_create_dim(nc_id, 'n_hor', n_hor)
        call nc_dump_var(nc_id, 'sqrt_lam', (/ 'n_hor' /), lam_sqrt(:), (/'long_name'/), (/'square root of the eigenvalues'/))
        call nc_dump_var(nc_id, 'lam', (/ 'n_hor' /), lam(:), (/'long_name'/), (/'eigenvalues of the spatial correlation matrix'/))
        call nc_dump_var(nc_id, 'P', (/'n_hor', 'n_hor'/), P(:,:), (/'long_name'/), (/'eigenvectors'/))

        ! JFM: also write B itself (first recalculate), for use in postprocessing
        write(*,*) "save_latlon_covariance: recalculating B matrix from eigenvectors/eigenvalues"
        ! P_diag is the diagonalized P, and P itself has the eigenvectors
        ! P (reconstituted) = P * P_diag * P.T
        allocate(B(n_hor,n_hor))
        call matmul_fast(P, P_diag, B) ! B <- P * P_diag
        ! since P_diag will not be used any more, use it to hold P.T
        P_diag = transpose(P)
        call matmul_fast(B, P_diag, P) ! P <- B * P.T

        call nc_dump_var(nc_id, 'B', (/'n_hor', 'n_hor'/), P(:,:), (/'long_name'/), (/'Horizontal correlation matrix'/))

        deallocate( B )

        call nc_close(nc_id)
        ! restore global compression setting

        deallocate ( lam, lam_sqrt )
        deallocate ( P, P_diag )

    end subroutine save_latlon_covariance

    subroutine make_positive_semidef(lam)
        real, intent(inout) :: lam(:)
        integer              :: i, n, n_neg, status

        if (min_eigval .gt. 1.0E-10) then
            min_eigval = min_eigval * minval((/ 1.0, maxval(lam) /))
        end if

        write(*,'("Maximum eigenvalue = ", e10.3, ", minimum eigenvalue = ", e10.3)') maxval(lam), minval(lam)

        n_neg = 0
        n = size(lam,1)
        do i=1,n
            if (lam(i) .lt. min_eigval) then
                lam(i) = min_eigval
                n_neg = n_neg + 1
            end if
        end do
        if (n_neg .gt. 0) write(*,'("Set ", i5, " eigenvalues to ", f15.11)') n_neg, min_eigval
    end subroutine make_positive_semidef

end module covar_latlon
