!
! Miscallaneous tools
!
!  subroutine dist
!     compute distance between points on the globe
!  subroutine dist_corr_fun
!     calculate a distance-dependent correlation function
!  subroutine eigen_symm
!     calculate eigenvectors/-values of symmetric matrix
!  function gasdev
!     Calculate normally distributed deviate with zero mean
!     and unit variance (from numerical recipes)
!  function ran_n
!     random number generator
!  subroutine check_dir
!     check if the directory for a file exists, and if not, create it
!  subroutine collapse_loop
!     collapse nested iterators into a single iterator
!
!---------------------------------------------------------------

module misctools
  implicit none
  private
  public :: dist
  public :: check_dir

  real, parameter :: pi=3.14159265358979
  real, parameter ::  ae = 6.371e6

contains

  subroutine check_dir(file_name)
    !---------------------------------------------------------------
    ! Check if the directory for a file exists or not, and if not, create it
    !---------------------------------------------------------------
    ! I/O
    character(len=*), intent(in)    :: file_name
    ! Local
    character(len=512)              :: dir_name
    logical                         :: dir_exists
    integer                         :: sep_pos
    ! Begin

    sep_pos = index(file_name, '/', .true.)
    ! If the file refers to the current folder or the root folder, no need to do anything more
    if (sep_pos .gt. 1) then
        dir_name = file_name(1:sep_pos-1)
        inquire(file=trim(dir_name)//'/.', exist=dir_exists)
        if (.not. dir_exists) call system('mkdir -p '//trim(dir_name))
    end if

  end subroutine check_dir

  subroutine dist( x1g, y1g, x2g, y2g, ddg )
    !---------------------------------------------------------------
    !  Compute distance of two points on the globe.
    !
    !  This formula is correct, also for points close together.
    !
    !  Input:
    !   x1g, y1g : longitude and latitude of first point  (degrees)
    !   x2g, y2g : longitude and latitude of second point (degrees)
    !
    !  Output:
    !   ddg      : distance between these points          (km)
    !----------------------------------------------------------------

    ! --- in/out ----------------------------------------------

    real, intent(in)               :: x1g, x2g, y1g, y2g
    real, intent(out)              :: ddg

    ! --- local -----------------------------------------------

    real                           :: x1, x2, y1, y2, dd
    real                           :: cc, dx, dy, dx2, dy2, dd2

    ! --- begin -----------------------------------------------

    cc = pi / 180.0
    !
    x1 = x1g * cc
    x2 = x2g * cc
    y1 = y1g * cc
    y2 = y2g * cc
    !
    !      print *,'x1,x2,y1,y2', x1,x2,y1,y2
    dy = 0.5 * ( y2 - y1 )
    dy = sin(dy)
    dy2 = dy * dy
    !
    dx = 0.5 * ( x2 - x1 )
    dx = sin(dx)
    dx2 = dx * dx * cos(y1) * cos(y2)
    !
    !     print *,'dx,dy,dx2,dy2',dx,dy,dx2,dy2
    dd2 = dx2 + dy2
    dd = sqrt(dd2)
    dd = 2.0 * asin(dd)
    !
    ddg = dd / cc
    ddg = (ddg * 2 * pi * 0.001 * ae) / 360
    !     print *,'dd2,dd,ddg', dd2,dd,ddg
    !
  end subroutine dist


end module misctools
