#!/usr/bin/env python

"""
This scripts creates a multipage pdf from one footprint file. The assumption is that the file contains footprints of multiple successive observations at the same site. It will plot on every page the 2D footprint of each observation (i.e. sum of the 3D footprint along its time axis).

Parameters:

    - name of the footprint file
    - vmin/vmax : the min/max values (set to None to have them automatically set, but it will vary from one figure to the other)
    - log : True for log scale, False otherwise

A pdf file is written out with the same name (and in the same folder) as the footprint file (with the extension changed from h5 to pdf)
"""

from h5py import File
from mpl_toolkits.basemap import Basemap
from matplotlib.backends.backend_pdf import PdfPages
from numpy import zeros, log, exp
from matplotlib.pyplot import clf

#================================
# Settings:
filename = 'filename.h5'
vmin, vmax = -10, 0
logscale=True

#================================
fid = File(filename, 'r')
pp = PdfPages(fid.filename.replace('h5', 'pdf'))

lons = fid['latitudes'][:]
lats = fid['longitudes'][:]

# Create map (lats/lons are center of grid box, so we need to add/remove half intervals
dlat = lats[1]-lats[0]
dlon = lons[1]-lons[0]
m1 = Basemap(
        llcrnrlat=lats.min()-dlat/2.,
        llcrnrlon=lons.min()-dlon/2.,
        urcrnrlat=lats.max()+dlat/2.,
        urcrnrlon=lons.max()+dlon/2.,
        resolution='l'
)

# Get the obs list (all but lats/lons:
obslist = sorted([x for x in fid.keys() if not x in ['latitudes', 'longitudes']])

# Loop over the observations (they should be sorted by time):
for obs in obslist:
    print obs
    # Create 2d array to store the aggregated footprint:
    fp2d = zeros((lats.size, lons.size))
    
    # Aggregate all the time intervals:
    for interv in fid[obs].keys():
        ilats = fid[obs][interv]['ilats'][:]
        ilons = fid[obs][interv]['ilons'][:]
        fp2d[ilats, ilons] += fid[obs][interv]['resp'][:]

    # Plot, in log scale
    if logscale :
        m1.imshow(log(fp2d), vmin=vmin, vmax=vmax)
    else :
        m1.imshow(fp2d, vmin=exp(vmin), vmax=exp(vmax))
    m1.drawcoastlines()
    m1.colorbar()
    pp.savefig()
    clf()
pp.close()
fid.close()
